from . import api
from . import tools

from .api import LSB_MWS
from .tools import pngify
from .version import VERSION

from PIL import Image


__version__ = VERSION
__author__ = 'thenonproton@pm.me'
